#include "StdAfx.h"
#include "ImageGenerator.h"

// FD300 title surface info ...
// ... Title mixer4 input (for external mixer usage - 30MBytes/sec PCI bandwith)
//GUID CImageGenerator::guidTitleTarget = CLSID_SLTMTitleNoPremultServerTargetMgr;
// ... Title mixer4 input (for internal mixer usage - minimum PCI bandwith)
GUID CImageGenerator::guidTitleTarget = CLSID_SLTMTitleServerTargetMgr;
// ... Artist/Foreground mixer4 input (20MBytes/sec PCI bandwith)
//GUID CImageGenerator::guidTitleTarget = CLSID_SLTMTitleArtServerTargetMgr;
// ... Scene/Background mixer4 input (20MBytes/sec PCI bandwith)
//GUID CImageGenerator::guidTitleTarget = CLSID_SLTMTitleBackServerTargetMgr;
// ... FD300 board index (starting from '0')
long CImageGenerator::nBoardIndex = 0;

HRESULT CImageGenerator::Init(RECT rect, long nZOrder)
{
	ATLASSERT((rect.left & 15)==0);		// must be aligned to 16
	ATLASSERT((rect.right & 15)==0);	// must be aligned to 16
	ATLASSERT((rect.top & 1)==0);		// must be aligned to 2
	ATLASSERT((rect.bottom & 1)==0);	// must be aligned to 2
	ATLASSERT(nZOrder != 0);
	// create low-level title engine
	HRESULT hres = m_pITitler.CoCreateInstance(CLSID_SLTitleTarget);
	if( FAILED(hres) )  return hres;
	// Connect title engine to display device (for external mixer usage)
	hres = m_pITitler->ConnectLocalHost(guidTitleTarget, nBoardIndex);
	if( FAILED(hres) )  return hres;
	// obtain target info
	CTargetParams target_params;
	hres = m_pITitler->GetTargetParams(&target_params);
	if( FAILED(hres) )  return hres;
	// upper field first if starting time of upper field lower then starting time of lower field
	m_bUpperField = target_params.m_UpperFieldTimerShift < target_params.m_LowerFieldTimerShift;
	// create COM-object for application alive checking
	hres = m_pIHolder.CoCreateInstance(CLSID_SLResourceHolder);
	if( FAILED(hres) )  return hres;
	// obtain region creator interface from title engine
	CComQIPtr<ISLTitleSurface> pISurface(m_pITitler);
	if( pISurface==NULL )  return E_NOINTERFACE;
	// create region with 2MB fifo size
	CComPtr<IUnknown> pIUnk;
	hres = pISurface->CreateTitleRectRegion(m_pIHolder, 2*1024*1024, 2*1024*1024, &pIUnk);
	if( FAILED(hres) )  return hres;
	hres = pIUnk->QueryInterface(&m_pIRegion);
	if( FAILED(hres) )  return hres;
	// init region (area, Z-order and so on)
	CRegionInfo region_info;
	region_info.x0 = rect.left;
	region_info.y0 = rect.top;
	region_info.x1 = rect.right;
	region_info.y1 = rect.bottom;
	region_info.z = nZOrder;
	region_info.ShowLastHalfShot = m_bShowLastHalfShot ? TRUE : FALSE;
	region_info.alfa = 0;
	region_info.ToAlfa = 255;
	region_info.FromAlfaShot = 0;
	region_info.ToAlfaShot = m_nFadeInCounter;
	region_info.CurrShot = 0;
	CSLTitleRegionPlacementFlds set_bits = (CSLTitleRegionPlacementFlds)
		(SLTRG_XYPosition | SLTRG_ZOrder | SLTRG_Alfa | SLTRG_ShowLastHalfShot | SLTRG_CurrShot | SLTRG_ToAlfa);
	CSLTitleRegionPlacementFlds return_bits = (CSLTitleRegionPlacementFlds)0;
	hres = m_pIRegion->SetParameters(set_bits, &return_bits, &region_info);
	if( FAILED(hres) )  return hres;
	// all right - ready to show
	m_nFrameIndex = 0;
	m_nDX = rect.right - rect.left;
	m_nDY = rect.bottom - rect.top;
	m_nImageSize = m_nDY * m_nDX * 4;
	return S_OK;
}

HRESULT CImageGenerator::Release(bool no_check)
{
	if( m_bNeedToWork ) {
		if( !no_check )  return E_ACCESSDENIED;
		m_bNeedToWork = false;
	}
	// close region
	m_pIRegion = NULL;
	// close COM-object for application alive checking
	m_pIHolder = NULL;
	// all right
	return S_OK;
}

HRESULT CImageGenerator::Work()
{
	HRESULT hres = S_OK;
	bool bWasStarted = false;
	while( m_bNeedToWork ) {
		// obtain pointer to new image
		RGBQUAD *ptr;
		HRESULT hres = GetFreeSpace4Image(ptr);
		if( FAILED(hres) )  break;
		if( hres==S_FALSE ) {
			// no free space into fifo
			if( !bWasStarted ) {
				// start playback
				hres = m_pIRegion->SetActive(TRUE, TRUE, 0);
				if( FAILED(hres) )  break;
				bWasStarted = true;
			} else {
				// wait some time
				Sleep(33);
			}
		} else { // all right - ready for next image
			ATLASSERT(hres==S_OK);
			// increment image frame index
			++m_nFrameIndex;
			// draw next image
			hres = DrawNextImage(ptr, m_nFrameIndex);
			if( FAILED(hres) )  break;
			// add the image into fifo
			hres = ConfirmImageInFifo();
			if( FAILED(hres) )  break;
		}
	}
	if( hres==S_OK )  hres = FadeOut();
	// hide the region
	m_pIRegion->SetActive(FALSE, FALSE, 0);
	return hres;
}

HRESULT CImageGenerator::GetFreeSpace4Image(RGBQUAD* &image)
{
	// obtain free space in region fifo
	BYTE *buf=NULL;  DWORD size;
	HRESULT hres = m_pIRegion->GetShotBuff(m_nImageSize, m_nImageSize, &buf, &size, 0);
	if( FAILED(hres) )  return hres;
	if( buf==NULL ) // no free space - need wait
		return S_FALSE;
	// all right
	image = (RGBQUAD*)buf;
	return S_OK;
}

HRESULT CImageGenerator::ConfirmImageInFifo()
{
	// confirm valid image data
	CTitleShotInfoEx shot_info;
	shot_info.SequenceNumber = 0;
	shot_info.DataType = TitleShotDataType_RGBA;
	shot_info.ShotNumber = m_nFrameIndex;
	shot_info.dx = m_nDX;
	shot_info.dy = m_nDY;
	shot_info.alfa = 255;
	shot_info.ShowLastFullShot = TRUE;
	return m_pIRegion->ConfirmShotSizeEx(0, m_nImageSize, &shot_info);
}

HRESULT CImageGenerator::FadeOut()
{
	// fill info for fade-out alpha interpolation
	CRegionInfo region_info;
	region_info.FromAlfaShot = 0;
	region_info.ToAlfaShot = -m_nFadeOutCounter;
	region_info.ToAlfa = 0;
	// make fade-out
	CSLTitleRegionPlacementFlds set_bits = SLTRG_ToAlfa;
	CSLTitleRegionPlacementFlds return_bits = (CSLTitleRegionPlacementFlds)0;
	HRESULT hres = m_pIRegion->SetParameters(set_bits, &return_bits, &region_info);
	if( FAILED(hres) )  return hres;
	// wait fade-out time
	Sleep(40*m_nFadeOutCounter);
	// all right
	return S_OK;
}
